# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fingerprint related models."""

from dataclasses import dataclass
from typing import Any, Literal, Self, TypedDict


class FingerprintNotFoundResponse(TypedDict):
    """TODO."""

    type: Literal["fingerprint-not-found"] = "fingerprint-not-found"


class FingerprintNotFoundError(Exception):
    """TODO."""


@dataclass
class Fingerprint:
    """Represents a fingerprint.

    See [Fingerprints](https://app.voker.ai/docs/concepts/events#fingerprints) for more information.
    """

    fingerprint_id: str

    @classmethod
    def deserialize(cls, data: dict[str, Any]) -> Self:
        """TODO."""
        return cls(fingerprint_id=data["fingerprint_id"])



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fingerprint related payloads and responses."""

from typing import NotRequired, Required, TypedDict


class CreateFingerprintPayload(TypedDict):
    """Create a fingerprint.

    See [Fingerprints](https://app.voker.ai/docs/concepts/events#fingerprints) for more information.
    """

    language: Required[str]
    language_version: Required[str]
    system: Required[str]

    openai_sdk_version: NotRequired[str]
    anthropic_sdk_version: NotRequired[str]
    gemini_sdk_version: NotRequired[str]
    voker_sdk_version: NotRequired[str]
    all_packages: NotRequired[dict[str, str]]
    git_repository_url: NotRequired[str]
    git_branch: NotRequired[str]
    git_commit_hash: NotRequired[str]
    git_commit_date: NotRequired[str]



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Directly call the Voker API.

See `voker.api.client.ApiClient` for a higher-level client that wraps this API client.
"""

from __future__ import annotations

import importlib.metadata
import os
import platform
import subprocess
import sys
from typing import Any, Optional, cast

import requests

from voker.schema.api import (
    CreateAgentPayload,
    CreateAgentVersionPayload,
    CreateEventPayload,
    CreateFingerprintPayload,
    CreatePersonPayload,
    EventProperties,
    UpdatePersonPayload,
)
from voker.schema.base import (
    Agent,
    AgentNotFoundError,
    AgentNotFoundResponse,
    AgentVersion,
    AgentVersionNotFoundError,
    AgentVersionNotFoundResponse,
    Fingerprint,
    FingerprintNotFoundError,
    FingerprintNotFoundResponse,
    Person,
    PersonNotFoundError,
    PersonNotFoundResponse,
)


class ApiClient:
    """Voker API Client.

    See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.api.client.ApiClient)
    """

    def __init__(
        self,
        *,
        base_url: str = "https://evals.voker.ai",
        api_key: Optional[str] = None,
    ) -> None:
        self._base_url = base_url

        if api_key is None:
            api_key = os.getenv("VOKER_API_KEY")

        self._api_key = api_key

    @property
    def events(self) -> ApiClient_Events:
        """Access [events](https://app.voker.ai/docs/concepts/events#events) related endpoints."""
        return ApiClient_Events(self)

    @property
    def people(self) -> ApiClient_People:
        """Access [people](https://app.voker.ai/docs/concepts/people) related endpoints."""
        return ApiClient_People(self)

    @property
    def fingerprints(self) -> ApiClient_Fingerprints:
        """Access [fingerprints](https://app.voker.ai/docs/concepts/events#fingerprints) related endpoints."""
        return ApiClient_Fingerprints(self)

    @property
    def agents(self) -> ApiClient_Agents:
        """Access [agents](https://app.voker.ai/docs/concepts/agents#agents) related endpoints."""
        return ApiClient_Agents(self)

    @property
    def agent_versions(self) -> ApiClient_AgentVersions:
        """Access [agent versions](https://app.voker.ai/docs/concepts/agents#agent-versions) related endpoints."""
        return ApiClient_AgentVersions(self)

    def _request(self, method: str, path: str, **kwargs) -> requests.Response:
        url = f"{self._base_url}{path}"
        headers = kwargs.pop("headers", {})

        if self._api_key is not None:
            headers["Authorization"] = f"Bearer {self._api_key}"

        return requests.request(method, url, headers=headers, **kwargs)


class _ApiClient_BaseSubPath:

    def __init__(self, voker_client: ApiClient) -> None:
        self._voker_client = voker_client


class ApiClient_Events(_ApiClient_BaseSubPath):
    """Access the [events](https://app.voker.ai/docs/concepts/events) API endpoints.

    See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.api.client.ApiClient_Events)
    """

    def create(
        self,
        *,
        event_name: str,
        properties: EventProperties,
        fingerprint_id: str,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
    ) -> None:
        """Create a new [event](https://app.voker.ai/docs/concepts/events).

        See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.api.client.ApiClient_Events.create)

        Parameters
        ----------
        event_name: str
        properties: EventProperties
        fingerprint_id: str
            The [fingerprint](https://app.voker.ai/docs/concepts/events#fingerprints) of the client that generated the event.
        agent: Optional[str]
            The name of the [agent](https://app.voker.ai/docs/concepts/agents#agents) associated with the event.
            If this agent does not exist, it will be created. If an agent is not supplied, a default agent will be used.
        agent_version: Optional[str]
            The name of the [agent version](https://app.voker.ai/docs/concepts/agents#agent-versions) associated with the event.
            If this agent version does not exist, it will be created. If an agent version is not supplied, a default version will be used.
        person: Optional[str]
            The unique identifier of the [person](https://app.voker.ai/docs/concepts/people) associated with the event.
            If the person does not exist, it will be created.
        session: Optional[str]
            The unique identifier of the [session](https://app.voker.ai/docs/concepts/events#event-sessions) associated with the event.
            If the session does not exist, it will be created.
        """
        payload = CreateEventPayload(
            event_name=event_name,
            fingerprint_id=fingerprint_id,
            properties=properties,
        )

        if agent is not None:
            payload["agent"] = agent

        if agent_version is not None:
            payload["agent_version"] = agent_version

        if person is not None:
            payload["person"] = person

        if session is not None:
            payload["session"] = session

        response = self._voker_client._request(
            "POST",
            "/api/v1/events",
            json=payload,
        )

        data = response.json()

        if response.status_code == 404:
            data = cast(FingerprintNotFoundResponse, data)
            raise FingerprintNotFoundError()

        response.raise_for_status()


class ApiClient_People(_ApiClient_BaseSubPath):
    """Access the [people](https://app.voker.ai/docs/concepts/people) API endpoints.

    See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.api.client.ApiClient_People)
    """

    def get(self, *, person_id: str) -> Person:
        """Retrieve a [person](https://app.voker.ai/docs/concepts/people)."""
        response = self._voker_client._request("GET", f"/api/v1/people/{person_id}")

        data = response.json()

        if response.status_code == 404:
            data = cast(PersonNotFoundResponse, data)
            raise PersonNotFoundError()

        response.raise_for_status()

        return Person.deserialize(data)

    def create(self, *, person_id: str, properties: dict[str, Any]) -> Person:
        """Create a person.

        See [Creating a Person](https://app.voker.ai/docs/concepts/people#creating-a-person) for more information.

        Parameters
        ----------
        person_id: str
            Unique identifier for the person.
        properties: dict[str, Any]
        """
        response = self._voker_client._request(
            "PUT",
            "/api/v1/people",
            json=CreatePersonPayload(person_id=person_id, properties=properties),
        )

        response.raise_for_status()

        data = response.json()

        return Person.deserialize(data)

    def update(self, *, person_id: str, properties: dict[str, Any]) -> Person:
        """Update a person."""
        response = self._voker_client._request(
            "PUT",
            f"/api/v1/people/{person_id}",
            json=UpdatePersonPayload(properties=properties),
        )

        data = response.json()

        if response.status_code == 404:
            data = cast(PersonNotFoundResponse, data)
            raise PersonNotFoundError()

        response.raise_for_status()

        return Person.deserialize(data)


def _git(cmd: str) -> str:
    return (
        subprocess.check_output(cmd.split(), stderr=subprocess.DEVNULL).decode().strip()
    )


class ApiClient_Fingerprints(_ApiClient_BaseSubPath):
    """Access [fingerprints](https://app.voker.ai/docs/concepts/events#fingerprints) related endpoints.

    See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.api.client.ApiClient_Fingerprints)
    """

    def create(self) -> Fingerprint:
        """Create a fingerprint.

        See [Fingerprints](https://app.voker.ai/docs/concepts/events#fingerprints) for more information.
        """
        payload = CreateFingerprintPayload(
            language="python",
            language_version=sys.version,
            system=platform.platform(),
        )

        try:
            payload["openai_sdk_version"] = importlib.metadata.version("openai")
        except importlib.metadata.PackageNotFoundError:
            ...

        try:
            payload["anthropic_sdk_version"] = importlib.metadata.version("anthropic")
        except importlib.metadata.PackageNotFoundError:
            ...

        try:
            payload["gemini_sdk_version"] = importlib.metadata.version("google.genai")
        except importlib.metadata.PackageNotFoundError:
            ...

        try:
            payload["voker_sdk_version"] = importlib.metadata.version("voker")
        except importlib.metadata.PackageNotFoundError:
            ...

        payload["all_packages"] = {
            dist.metadata["Name"]: dist.version
            for dist in importlib.metadata.distributions()
        }

        try:
            payload["git_repository_url"] = _git("git config --get remote.origin.url")
            payload["git_branch"] = _git("git rev-parse --abbrev-ref HEAD")
            payload["git_commit_hash"] = _git("git rev-parse HEAD")
            payload["git_commit_date"] = _git("git show -s --format=%ci HEAD")
        except:
            ...

        response = self._voker_client._request(
            "PUT",
            "/api/v1/fingerprints",
            json=payload,
        )

        response.raise_for_status()

        data = response.json()

        return Fingerprint.deserialize(data)


class ApiClient_Agents(_ApiClient_BaseSubPath):
    """Access [agents](https://app.voker.ai/docs/concepts/agents#agents) related endpoints.

    See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.api.client.ApiClient_Agents)
    """

    def create(self, *, agent_name: str, description: Optional[str] = None) -> Agent:
        """Create an agent.

        See [Creating an Agent](https://app.voker.ai/docs/concepts/agents#creating-agent) for more information.

        Parameters
        ----------
        agent_name: str
            Unique identifier for the agent.
        description: Optional[str]
        """
        payload = CreateAgentPayload(agent_name=agent_name)

        if description is not None:
            payload["description"] = description

        response = self._voker_client._request(
            "PUT",
            "/api/v1/agents",
            json=payload,
        )

        response.raise_for_status()

        data = response.json()

        return Agent.deserialize(data)

    def get(self, *, agent_name: str) -> Agent:
        """Get an agent."""
        response = self._voker_client._request("GET", f"/api/v1/agents/{agent_name}")

        data = response.json()

        if response.status_code == 404:
            data = cast(AgentNotFoundResponse, data)
            raise AgentNotFoundError()

        response.raise_for_status()

        return Agent.deserialize(data)


class ApiClient_AgentVersions(_ApiClient_BaseSubPath):
    """Access [agent versions](https://app.voker.ai/docs/concepts/agents#agent-versions) related endpoints.

    See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.api.client.ApiClient_AgentVersions)
    """

    def create(
        self,
        *,
        agent_name: str,
        agent_version_name: str,
        description: Optional[str] = None,
    ) -> AgentVersion:
        """Create an agent version.

        See [Creating an Agent Version](https://app.voker.ai/docs/concepts/agents#creating-agent-version) for more information.

        Parameters
        ----------
        agent_name: str
            The name of the existing agent that this version belongs to.
        agent_version_name: str
            Unique identifier for the agent version within its agent.
        description: Optional[str]
        """
        payload = CreateAgentVersionPayload(agent_version_name=agent_version_name)

        if description is not None:
            payload["description"] = description

        response = self._voker_client._request(
            "PUT",
            f"/api/v1/agents/{agent_name}/agent-versions",
            json=payload,
        )

        data = response.json()

        if response.status_code == 404:
            data = cast(AgentNotFoundResponse, data)
            raise AgentNotFoundError()

        response.raise_for_status()

        return AgentVersion.deserialize(data)

    def get(self, *, agent_name: str, agent_version_name: str) -> AgentVersion:
        """Get an agent version."""
        response = self._voker_client._request(
            "GET",
            f"/api/v1/agents/{agent_name}/agent-versions/{agent_version_name}",
        )

        data = response.json()

        if response.status_code == 404:
            data = cast(AgentNotFoundResponse | AgentVersionNotFoundResponse, data)

            if data["detail"]["type"] == "agent-not-found":
                raise AgentNotFoundError()
            elif data["detail"]["type"] == "agent-version-not-found":
                raise AgentVersionNotFoundError()

        response.raise_for_status()

        return AgentVersion.deserialize(data)



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implement Voker tracking example for Gemini.

[See Here](https://app.voker.ai/docs/integrations/providers/gemini).
"""

from __future__ import annotations

from collections.abc import AsyncIterator, Iterator
from typing import TYPE_CHECKING, Any, Optional, overload

from voker.client import VokerClient

if TYPE_CHECKING:
    from google.genai import _interactions as genai_interactions
    from google.genai import models as genai_models
    from google.genai import types as genai_types
    from google.genai._interactions import resources as genai_interactions_resources
    from google.genai._interactions import types as genai_interactions_types

try:
    import google
    from google import genai
except ImportError:
    raise ModuleNotFoundError


class Client(genai.Client):
    """Gemini wrapper that automatically creates llm events in Voker."""

    @overload
    def __init__(
        self,
        voker_client: Optional[VokerClient] = None,
        base_genai_client: Optional[genai.Client] = None,
        *,
        vertexai: Optional[bool] = None,
        api_key: Optional[str] = None,
        credentials: Optional[google.auth.credentials.Credentials] = None,
        project: Optional[str] = None,
        location: Optional[str] = None,
        debug_config: Optional[genai.client.DebugConfig] = None,
        http_options: Optional[
            genai_types.HttpOptions | genai_types.HttpOptionsDict
        ] = None,
    ) -> None: ...

    def __init__(
        self,
        voker_client: Optional[VokerClient] = None,
        base_genai_client: Optional[genai.Client] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if voker_client is not None:
            self.__voker_client = voker_client
        else:
            self.__voker_client = VokerClient()

        if base_genai_client is None:
            base_genai_client = self
            self.__base_genai_client = base_genai_client

        if (base_models := getattr(base_genai_client, "_models", None)) is not None:
            self._models = Models(self.__voker_client, base_models)

        if (
            base_interactions := getattr(base_genai_client, "interactions", None)
        ) is not None:
            self._nextgen_client_instance

            self._nextgen_client.interactions = Interactions(
                self.__voker_client,
                base_interactions,
            )

    @property
    def aio(self) -> AsyncClient:
        """Pass through to the underlying AsyncClient instance."""
        return AsyncClient(
            self.__voker_client,
            self.__base_genai_client._aio,
        )


class Models:
    """Send calls to the models endpoints to Voker."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_models: genai_models.Models,
    ) -> None:
        self.__voker_client = voker_client
        self.__base_models = base_models

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying Models instance."""
        return getattr(self.__base_models, name)

    def generate_content(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> genai_types.GenerateContentResponse:
        """Send a content generation request to Voker as an llm event.

        See `google.genai.Client.models.generate_content`.
        """
        response = self.__base_models.generate_content(**kwargs)

        self.__voker_client.events.create(
            event_name="llm",
            properties={
                "api": "gemini-models",
                "inputs": kwargs,
                "output": response.model_dump(mode="json"),
            },
            agent=agent,
            agent_version=agent_version,
            person=person,
            session=session,
        )

        return response

    def generate_content_stream(
        self,
        **kwargs,
    ) -> Iterator[genai_types.GenerateContentResponse]:
        """Send a streaming content generation request to Voker as an llm event.

        See `google.genai.Client.models.generate_content_stream`.
        """
        for chunk in self.__base_models.generate_content_stream(**kwargs):
            yield chunk

        # TODO: accumulate


class Interactions:
    """Send calls to the interactions endpoints to Voker."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_interactions: genai_interactions_resources.InteractionsResource,
    ) -> None:
        self.__voker_client = voker_client
        self.__base_interactions = base_interactions

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying Interactions instance."""
        return getattr(self.__base_interactions, name)

    def create(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> (
        genai_interactions_types.Interaction
        | genai_interactions.Stream[genai_interactions_types.InteractionSSEEvent]
    ):
        """Send a create interaction request to Voker as an llm event.

        See: `google.genai.Client.interactions.create`
        """
        from google.genai._interactions.types import Interaction

        stream = kwargs.get("stream", False)

        interaction = self.__base_interactions.create(**kwargs)

        if not stream:
            self.__voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "gemini-interactions",
                    "inputs": kwargs,
                    "output": interaction,
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

            return interaction

        # TODO: stream and accumulate...

        return interaction


class AsyncClient:
    """Async Gemini client wrapper that automatically creates llm events in Voker."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_genai_async_client: genai.client.AsyncClient,
    ) -> None:
        self.__voker_client = voker_client
        self.__base_genai_async_client = base_genai_async_client

        if (
            base_models := getattr(self.__base_genai_async_client, "models", None)
        ) is not None:
            self._models = AsyncModels(voker_client, base_models)

        if (
            base_interactions := getattr(base_genai_async_client, "interactions", None)
        ) is not None:
            self._nextgen_client.interactions = AsyncInteractions(
                self.__voker_client,
                base_interactions,
            )

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying AsyncClient object."""
        return getattr(self.__base_genai_async_client, name)


class AsyncModels:
    """Send async calls to the models endpoints to Voker as llm events."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_models: genai_models.AsyncModels,
    ) -> None:
        self.__voker_client = voker_client
        self.__base_models = base_models

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying AsyncModels object."""
        return getattr(self.__base_models, name)

    async def generate_content(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> genai_types.GenerateContentResponse:
        """Send an async content generation request to Voker as an llm event.

        See: `google.genai.client.AsyncClient.models.generate_content`
        """
        response = await self.__base_models.generate_content(**kwargs)

        self.__voker_client.events.create(
            event_name="llm",
            properties={
                "api": "gemini-models",
                "inputs": kwargs,
                "output": response.model_dump(mode="json"),
            },
            agent=agent,
            agent_version=agent_version,
            person=person,
            session=session,
        )

        return response

    async def generate_content_stream(
        self,
        **kwargs,
    ) -> AsyncIterator[genai_types.GenerateContentResponse]:
        """Send an async streaming content generation request to Voker as an llm event.

        See: `google.genai.client.AsyncClient.models.generate_content_stream`
        """
        stream = await self.__base_models.generate_content_stream(**kwargs)

        async for chunk in stream:
            yield chunk

        # TODO: accumulate


class AsyncInteractions:
    """Send async calls to the interactions endpoints to Voker."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_interactions: genai_interactions_resources.AsyncInteractionsResource,
    ) -> None:
        self.__voker_client = voker_client
        self.__base_interactions = base_interactions

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying Interactions instance."""
        return getattr(self.__base_interactions, name)

    async def create(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> (
        genai_interactions_types.Interaction
        | genai_interactions.AsyncStream[genai_interactions_types.InteractionSSEEvent]
    ):
        """Send a, async create interaction request to Voker as an llm event.

        See: `google.genai.client.AsyncClient.interactions.create`
        """
        from google.genai._interactions.types import Interaction

        stream = kwargs.get("stream", False)

        interaction = await self.__base_interactions.create(**kwargs)

        if not stream:
            self.__voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "gemini-interactions",
                    "inputs": kwargs,
                    "output": interaction,
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

            return interaction

        # TODO: stream and accumulate...

        return interaction



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base models."""

from .agent_versions import (
    AgentVersion,
    AgentVersionNotFoundError,
    AgentVersionNotFoundResponse,
)
from .agents import Agent, AgentNotFoundError, AgentNotFoundResponse
from .fingerprints import (
    Fingerprint,
    FingerprintNotFoundError,
    FingerprintNotFoundResponse,
)
from .people import Person, PersonNotFoundError, PersonNotFoundResponse

__all__ = [
    "Agent",
    "AgentNotFoundError",
    "AgentNotFoundResponse",
    "AgentVersion",
    "AgentVersionNotFoundError",
    "AgentVersionNotFoundResponse",
    "Fingerprint",
    "FingerprintNotFoundError",
    "FingerprintNotFoundResponse",
    "Person",
    "PersonNotFoundError",
    "PersonNotFoundResponse",
]



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Agent version related models."""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Literal, Self, TypedDict


@dataclass
class AgentVersion:
    """Represents an agent version.

    See [Agent Versions](https://app.voker.ai/docs/concepts/agents#agent-versions) for more information.
    """

    agent_version_name: str
    """Unique identifier for the agent version within its agent."""

    description: str | None

    agent_version_number: int
    """Unique number assigned to the agent version within its agent."""

    created_at: datetime
    deprecated_at: datetime | None

    @classmethod
    def deserialize(cls, data: dict[str, Any]) -> Self:
        """TODO."""
        return cls(
            agent_version_name=data["agent_version_name"],
            description=data["description"],
            agent_version_number=data["agent_version_number"],
            created_at=datetime.fromisoformat(data["created_at"]),
            deprecated_at=(
                datetime.fromisoformat(data["deprecated_at"])
                if data["deprecated_at"] is not None
                else None
            ),
        )


class AgentVersionNotFoundResponse(TypedDict):
    """TODO."""

    type: Literal["agent-version-not-found"] = "agent-version-not-found"


class AgentVersionNotFoundError(Exception):
    """TODO."""



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implement Voker tracking example for Anthropic.

[See Here](https://app.voker.ai/docs/integrations/providers/anthropic).
"""

from __future__ import annotations

from collections.abc import Mapping
from functools import partial
from typing import TYPE_CHECKING, Any, Optional, overload

from voker.client import VokerClient

if TYPE_CHECKING:
    import httpx
    from anthropic import resources as anthropic_resources
    from anthropic import types as anthropic_types
    from anthropic.lib import streaming as anthropic_lib_streaming
    from anthropic.types import beta as anthropic_types_beta
    from anthropic.types import message as anthropic_types_message

try:
    import anthropic
    import httpx
except ImportError:
    raise ModuleNotFoundError

try:
    from anthropic import not_given
except ImportError:
    from anthropic import NOT_GIVEN as not_given


def _rm_not_given(
    data: dict[str, Any | anthropic.NotGiven | anthropic.Omit],
) -> dict[str, Any]:
    return {
        k: v
        for k, v in data.items()
        if not isinstance(v, (anthropic.NotGiven, anthropic.Omit))
    }


class Anthropic(anthropic.Anthropic):
    """Anthropic wrapper that automatically creates llm events in Voker."""

    @overload
    def __init__(
        self,
        voker_client: Optional[VokerClient] = None,
        base_anthropic_client: Optional[anthropic.Anthropic] = None,
        *,
        api_key: Optional[str] = None,
        auth_token: Optional[str] = None,
        base_url: Optional[str | httpx.URL] = None,
        timeout: Optional[float | anthropic.Timeout | anthropic.NotGiven] = not_given,
        max_retries: int = anthropic.DEFAULT_MAX_RETRIES,
        default_headers: Optional[Mapping[str, str]] = None,
        default_query: Optional[Mapping[str, object]] = None,
        # Configure a custom httpx client.
        # We provide a `DefaultHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        http_client: Optional[httpx.Client] = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None: ...

    def __init__(
        self,
        voker_client: Optional[VokerClient] = None,
        base_anthropic_client: Optional[anthropic.Anthropic] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        if voker_client is None:
            voker_client = VokerClient()

        if base_anthropic_client is None:
            base_anthropic_client = self

        # keep for testing access:
        self.__base_anthropic_client = base_anthropic_client
        self.__voker_client = voker_client

        if (
            base_messages := getattr(base_anthropic_client, "messages", None)
        ) is not None:
            self.messages = Messages(voker_client, base_messages)

        if (base_beta := getattr(base_anthropic_client, "beta", None)) is not None:
            self.beta = Beta(voker_client, base_beta)


class Beta:
    """Send calls to the beta endpoints to Voker."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_beta: anthropic_resources.beta.beta.Beta,
    ) -> None:
        self.__base_beta = base_beta

        if (base_messages := getattr(base_beta, "messages", None)) is not None:
            self.messages = BetaMessages(voker_client, base_messages)

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying Beta object."""
        return getattr(self.__base_beta, name)


class BetaMessages:
    """Send calls to the messages endpoints to Voker."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_messages: anthropic_resources.beta.messages.messages.Messages,
    ) -> None:
        self.__voker_client = voker_client
        self.__base_messages = base_messages

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying Messages object."""
        return getattr(self.__base_messages, name)

    def create(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> anthropic_types_beta.BetaMessage:
        """Send a message creation request to Voker as an llm event.

        See `anthropic.Anthropic.messages.create`.
        """
        to_stream = kwargs.get("stream", False)

        if not to_stream:
            message = self.__base_messages.create(**kwargs)

            self.__voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "anthropic-beta-messages",
                    "inputs": _rm_not_given(kwargs),
                    "output": message.model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

            return message

        api_request = partial(self.__base_messages.create, **kwargs)

        def gen():
            aggregator = _AnthropicBetaMessagesStreamAggregator(
                headers=httpx.Headers().update(kwargs.get("extra_headers")),
                output_format=kwargs.get("response_format", anthropic.not_given),
            )

            for chunk in api_request():
                aggregator.add_event(chunk)
                yield chunk

            self.__voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "anthropic-beta-messages",
                    "inputs": _rm_not_given(kwargs),
                    "output": aggregator.get_result().model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

        return gen()


class Messages:
    """Send calls to the messages endpoints to Voker."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_messages: anthropic_resources.messages.messages.Messages,
    ) -> None:
        self.__voker_client = voker_client
        self.__base_messages = base_messages

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying Messages object."""
        return getattr(self.__base_messages, name)

    def create(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> anthropic_types_message.Message:
        """Send a message creation request to Voker as an llm event.

        See `anthropic.Anthropic.messages.create`.
        """
        to_stream = kwargs.get("stream", False)

        if not to_stream:
            message = self.__base_messages.create(**kwargs)

            self.__voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "anthropic-messages",
                    "inputs": _rm_not_given(kwargs),
                    "output": message.model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

            return message

        api_request = self.__base_messages.create(**kwargs)

        def gen():
            aggregator = _AnthropicMessagesStreamAggregator()

            for chunk in api_request:
                aggregator.add_event(chunk)
                yield chunk

            self.__voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "anthropic-messages",
                    "inputs": _rm_not_given(kwargs),
                    # "output": message.model_dump(mode="json"),
                    "output": aggregator.get_result().model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

        return gen()


class AsyncAnthropic(anthropic.AsyncAnthropic):
    """Async Anthropic client wrapper that automatically creates llm events in Voker."""

    @overload
    def __init__(
        self,
        voker_client: Optional[VokerClient] = None,
        base_anthropic_client: Optional[anthropic.AsyncAnthropic] = None,
        *,
        api_key: Optional[str] = None,
        auth_token: Optional[str] = None,
        base_url: Optional[str | httpx.URL] = None,
        timeout: Optional[float | anthropic.Timeout | anthropic.NotGiven] = not_given,
        max_retries: int = anthropic.DEFAULT_MAX_RETRIES,
        default_headers: Optional[Mapping[str, str]] = None,
        default_query: Optional[Mapping[str, object]] = None,
        # Configure a custom httpx client.
        # We provide a `DefaultAsyncHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#asyncclient) for more details.
        http_client: Optional[httpx.AsyncClient] = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None: ...

    def __init__(
        self,
        voker_client: Optional[VokerClient] = None,
        base_anthropic_client: Optional[anthropic.AsyncAnthropic] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        if voker_client is None:
            voker_client = VokerClient()

        if base_anthropic_client is None:
            base_anthropic_client = self

        # keep for testing access:
        self.__base_anthropic_client = base_anthropic_client
        self.__voker_client = voker_client

        if (
            base_messages := getattr(base_anthropic_client, "messages", None)
        ) is not None:
            self.messages = AsyncMessages(voker_client, base_messages)

        if (base_beta := getattr(base_anthropic_client, "beta", None)) is not None:
            self.beta = AsyncBeta(voker_client, base_beta)


class AsyncBeta:
    """Async Beta client wrapper that automatically creates llm events in Voker."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_beta_client: Optional[anthropic.AsyncBeta] = None,
    ) -> None:
        if (base_messages := getattr(base_beta_client, "messages", None)) is not None:
            self.messages = AsyncBetaMessages(voker_client, base_messages)

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying AsyncBeta object."""
        return getattr(self.__base_beta_client, name)


class AsyncBetaMessages:
    """Async Beta Messages client wrapper that automatically creates llm events in Voker."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_messages: anthropic_resources.beta.messages.messages.AsyncMessages,
    ) -> None:
        self.__voker_client = voker_client
        self.__base_messages = base_messages

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying AsyncMessages object."""
        return getattr(self.__base_messages, name)

    async def create(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> anthropic_types_message.Message:
        """Send an async message creation request to Voker as an llm event.

        See `anthropic.Anthropic.messages.create`.
        """
        to_stream = kwargs.get("stream", False)

        if not to_stream:
            message = await self.__base_messages.create(**kwargs)

            self.__voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "anthropic-beta-messages",
                    "inputs": _rm_not_given(kwargs),
                    "output": message.model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

            return message

        api_request = self.__base_messages.create(**kwargs)

        async def gen():
            aggregator = _AnthropicBetaMessagesStreamAggregator(
                headers=httpx.Headers().update(kwargs.get("extra_headers")),
                output_format=kwargs.get("response_format", anthropic.not_given),
            )

            async for chunk in await api_request:
                aggregator.add_event(chunk)
                yield chunk

            self.__voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "anthropic-beta-messages",
                    "inputs": _rm_not_given(kwargs),
                    "output": aggregator.get_result().model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

        return gen()


class AsyncMessages:
    """Send async calls to the messages endpoints to Voker as llm events."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_messages: anthropic_resources.messages.messages.AsyncMessages,
    ) -> None:
        self.__voker_client = voker_client
        self.__base_messages = base_messages

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying AsyncMessages object."""
        return getattr(self.__base_messages, name)

    async def create(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> anthropic_types_message.Message:
        """Send an async message creation request to Voker as an llm event.

        See `anthropic.Anthropic.messages.create`.
        """
        to_stream = kwargs.get("stream", False)

        if not to_stream:
            message = await self.__base_messages.create(**kwargs)

            self.__voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "anthropic-messages",
                    "inputs": _rm_not_given(kwargs),
                    "output": message.model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

            return message

        api_request = self.__base_messages.create(**kwargs)

        async def gen():
            aggregator = _AnthropicMessagesStreamAggregator()

            async for chunk in await api_request:
                aggregator.add_event(chunk)
                yield chunk

            self.__voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "anthropic-messages",
                    "inputs": _rm_not_given(kwargs),
                    "output": aggregator.get_result().model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

        return gen()


class _AnthropicMessagesStreamAggregator:

    def __init__(self) -> None:
        self.current: anthropic_types.Message | None = None

    def add_event(self, chunk: anthropic_types.RawMessageStreamEvent) -> None:
        from anthropic.lib.streaming._messages import accumulate_event

        self.current = accumulate_event(event=chunk, current_snapshot=self.current)

    def get_result(self) -> anthropic_types.Message:
        assert self.current is not None
        return self.current


class _AnthropicBetaMessagesStreamAggregator:

    def __init__(
        self,
        headers: httpx.Headers,
        output_format: anthropic_types_beta.ResponseFormatT | anthropic.NotGiven,
    ) -> None:
        self.current: anthropic_types_beta.Message | None = None
        self.headers = headers
        self.output_format = output_format

    def add_event(self, chunk: anthropic_types_beta.RawMessageStreamEvent) -> None:
        from anthropic.lib.streaming._beta_messages import accumulate_event

        self.current = accumulate_event(
            event=chunk,
            current_snapshot=self.current,
            request_headers=self.headers,
            output_format=self.output_format,
        )

    def get_result(self) -> anthropic_types_beta.Message:
        assert self.current is not None
        return self.current



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""LLM provider wrapping to implement Voker tracking.

Do not auto import this module directly. Instead, import from specific provider modules, e.g.,
`voker.ai.provider_openai`, `voker.ai.provider_gemini`, or `voker.ai.provider_anthropic`. This way
users won't need to install providers they don't use.
"""



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Person related payloads and responses."""

from typing import Any, NotRequired, Required, TypedDict


class CreatePersonPayload(TypedDict):
    """Create a person.

    See [Creating a Person](https://app.voker.ai/docs/concepts/people#creating-a-person) for more information.
    """

    person_id: Required[str]
    properties: NotRequired[dict[str, Any]]


class UpdatePersonPayload(TypedDict):
    """Update a person."""

    properties: Required[dict[str, Any]]



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TODO."""

from __future__ import annotations

from typing import Any, Optional

from .api import ApiClient, EventQueue, EventWorker, Request
from .schema.api import EventProperties


class VokerClient:
    """Voker Client.

    See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.client.VokerClient)
    """

    def __init__(
        self,
        *,
        base_url: str = "https://evals.voker.ai",
        api_key: Optional[str] = None,
    ) -> None:
        self._api_client = ApiClient(base_url=base_url, api_key=api_key)
        self._queue = EventQueue()
        self._event_worker = EventWorker(self._api_client, self._queue)

    @property
    def events(self) -> VokerClient_Events:
        """Access [events](https://app.voker.ai/docs/concepts/events#events) related endpoints."""
        return VokerClient_Events(self._queue)

    @property
    def people(self) -> VokerClient_People:
        """Access [people](https://app.voker.ai/docs/concepts/people) related endpoints."""
        return VokerClient_People(self._queue)

    @property
    def agents(self) -> VokerClient_Agents:
        """Access [agents](https://app.voker.ai/docs/concepts/agents#agents) related endpoints."""
        return VokerClient_Agents(self._queue)

    @property
    def agent_versions(self) -> VokerClient_AgentVersions:
        """Access [agent versions](https://app.voker.ai/docs/concepts/agents#agent-versions) related endpoints."""
        return VokerClient_AgentVersions(self._queue)

    def close(self) -> None:
        """Close the Voker client and stop the event worker."""
        self._event_worker.close()


class _VokerClient_BaseSubPath:

    def __init__(self, queue: EventQueue) -> None:
        self._queue = queue


class VokerClient_Events(_VokerClient_BaseSubPath):
    """Access [events](https://app.voker.ai/docs/concepts/events#events) related endpoints.

    See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.client.VokerClient_Events)
    """

    def create(
        self,
        *,
        event_name: str,
        properties: EventProperties,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
    ) -> None:
        """Create a new [event](https://app.voker.ai/docs/concepts/events).

        See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.client.VokerClient_Events.create)

        Parameters
        ----------
        event_name: str
        properties: EventProperties
        agent: Optional[str]
            The name of the [agent](https://app.voker.ai/docs/concepts/agents#agents) associated with the event.
            If this agent does not exist, it will be created. If an agent is not supplied, a default agent will be used.
        agent_version: Optional[str]
            The name of the [agent version](https://app.voker.ai/docs/concepts/agents#agent-versions) associated with the event.
            If this agent version does not exist, it will be created. If an agent version is not supplied, a default version will be used.
        person: Optional[str]
            The unique identifier of the [person](https://app.voker.ai/docs/concepts/people) associated with the event.
            If the person does not exist, it will be created.
        session: Optional[str]
            The unique identifier of the [session](https://app.voker.ai/docs/concepts/events#event-sessions) associated with the event.
            If the session does not exist, it will be created.
        """
        self._queue.put(
            Request(
                "events",
                "create",
                event_name=event_name,
                properties=properties,
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )
        )


class VokerClient_People(_VokerClient_BaseSubPath):
    """Access [people](https://app.voker.ai/docs/concepts/people) related endpoints.

    See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.client.VokerClient_People)
    """

    def create(
        self,
        *,
        person_id: str,
        properties: Optional[dict[str, Any]] = None,
    ) -> None:
        """Create a person.

        See [Creating a Person](https://app.voker.ai/docs/concepts/people#creating-a-person) for more information.

        Parameters
        ----------
        person_id: str
            Unique identifier for the person.
        properties: dict[str, Any]
        """
        self._queue.put(
            Request("people", "create", person_id=person_id, properties=properties)
        )

    def update(self, *, person_id: str, properties: dict[str, Any]) -> None:
        """Update a person."""
        self._queue.put(
            Request("people", "update", person_id=person_id, properties=properties)
        )


class VokerClient_Agents(_VokerClient_BaseSubPath):
    """Access [agents](https://app.voker.ai/docs/concepts/agents#agents) related endpoints.

    See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.client.VokerClient_Agents)
    """

    def create(self, *, agent_name: str, description: Optional[str] = None) -> None:
        """Create an agent.

        See [Creating an Agent](https://app.voker.ai/docs/concepts/agents#creating-agent) for more information.

        Parameters
        ----------
        agent_name: str
            Unique identifier for the agent.
        description: Optional[str]
        """
        self._queue.put(
            Request("agents", "create", agent_name=agent_name, description=description)
        )


class VokerClient_AgentVersions(_VokerClient_BaseSubPath):
    """Access [agent versions](https://app.voker.ai/docs/concepts/agents#agent-versions) related endpoints.

    See [the reference documentation](https://app.voker.ai/docs/integrations/sdks/python#voker.client.VokerClient_AgentVersions)
    """

    def create(
        self,
        *,
        agent_name: str,
        agent_version_name: str,
        description: Optional[str] = None,
    ) -> None:
        """Create an agent version.

        See [Creating an Agent Version](https://app.voker.ai/docs/concepts/agents#creating-agent-version) for more information.

        Parameters
        ----------
        agent_name: str
            The name of the existing agent that this version belongs to.
        agent_version_name: str
            Unique identifier for the agent version within its agent.
        description: Optional[str]
        """
        self._queue.put(
            Request(
                "agent-versions",
                "create",
                agent_name=agent_name,
                agent_version_name=agent_version_name,
                description=description,
            )
        )



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Batch requests and send in the background to the Voker API.

This is useful to send events to the Voker API without adding latency to your application.
"""

import queue
import threading
from typing import TypeVar

from voker.schema.base.fingerprints import Fingerprint

from .client import ApiClient

T = TypeVar("T")


class Request:
    """TODO."""

    def __init__(self, path: str, operation: str, **kwargs) -> None:
        self.path = path
        self.operation = operation
        self.kwargs = kwargs


class EventQueue(queue.Queue[Request]):
    """Queue for batching events before sending them to the Voker API."""


_STOP = object()


class EventWorker(threading.Thread):
    """Background worker that processes events from an EventQueue and sends them to the Voker API."""

    def __init__(self, client: ApiClient, event_queue: EventQueue) -> None:
        super().__init__(target=self._process, daemon=True)

        self._client = client
        self._event_queue = event_queue

        self.start()

    def _process(self) -> None:
        fingerprint = self._client.fingerprints.create()

        while True:
            item = self._event_queue.get()

            if item == _STOP:
                break

            self._process_one(item, fingerprint)

    def _process_one(self, item: Request, fingerprint: Fingerprint) -> None:
        if item.path == "events":
            if item.operation == "create":
                self._client.events.create(
                    fingerprint_id=fingerprint.fingerprint_id,
                    **item.kwargs,
                )
        elif item.path == "people":
            if item.operation == "create":
                self._client.people.create(**item.kwargs)
            elif item.operation == "update":
                self._client.people.update(**item.kwargs)
        elif item.path == "agents":
            if item.operation == "create":
                self._client.agents.create(**item.kwargs)
        elif item.path == "agent-versions":
            if item.operation == "create":
                self._client.agent_versions.create(**item.kwargs)

        # oops...

    def close(self) -> None:
        """Stop the worker and wait for it to finish processing all events in the queue."""
        self._event_queue.put(_STOP)
        self.join()  # TODO: timeout?



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Person related models."""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Literal, Self, TypedDict


@dataclass
class Person:
    """Represents a person.

    See [People](https://app.voker.ai/docs/concepts/people) for more information.
    """

    person_id: str
    properties: dict[str, Any] | None
    created_at: datetime

    @classmethod
    def deserialize(cls, data: dict[str, Any]) -> Self:
        """TODO."""
        return cls(
            person_id=data["person_id"],
            properties=data["properties"],
            created_at=datetime.fromisoformat(data["created_at"]),
        )


class PersonNotFoundResponse(TypedDict):
    """TODO."""

    type: Literal["person-not-found"] = "person-not-found"


class PersonNotFoundError(Exception):
    """TODO."""



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implement Voker tracking example for OpenAI.

[See Here](https://app.voker.ai/docs/integrations/providers/openai).
"""

from __future__ import annotations

from collections.abc import Mapping
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Generator,
    Optional,
    overload,
)

from voker.client import VokerClient

if TYPE_CHECKING:
    import httpx
    from openai import _streaming as openai_streaming
    from openai import resources as openai_resources
    from openai.resources import responses as openai_responses
    from openai.types import chat as openai_types_chat
    from openai.types import responses as openai_types_responses


try:
    import openai
except ImportError:
    raise ModuleNotFoundError


def _rm_not_given(
    data: dict[str, Any | openai.NotGiven | openai.Omit],
) -> dict[str, Any]:
    return {
        k: v
        for k, v in data.items()
        if not isinstance(v, (openai.NotGiven, openai.Omit))
    }


class OpenAI(openai.OpenAI):
    """OpenAI wrapper that automatically creates llm events in Voker."""

    @overload
    def __init__(
        self,
        voker_client: Optional[VokerClient] = None,
        base_openai_client: Optional[openai.OpenAI] = None,
        *,
        api_key: Optional[str | Callable[[], str]] = None,
        organization: Optional[str] = None,
        project: Optional[str] = None,
        webhook_secret: Optional[str] = None,
        base_url: Optional[str | httpx.URL] = None,
        websocket_base_url: Optional[str | httpx.URL] = None,
        timeout: Optional[float | openai.Timeout | openai.NotGiven] = openai.not_given,
        max_retries: int = openai.DEFAULT_MAX_RETRIES,
        default_headers: Optional[Mapping[str, str]] = None,
        default_query: Optional[Mapping[str, object]] = None,
        # Configure a custom httpx client.
        # We provide a `DefaultHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        http_client: Optional[httpx.Client] = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
        **kwargs,
    ) -> None: ...

    def __init__(
        self,
        voker_client: Optional[VokerClient] = None,
        base_openai_client: Optional[openai.OpenAI] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if voker_client is None:
            voker_client = VokerClient()

        if base_openai_client is None:
            base_openai_client = self

        # keep for testing access:
        self.__base_openai_client = base_openai_client
        self.__voker_client = voker_client

        if (base_chat := getattr(base_openai_client, "chat", None)) is not None:
            self.chat = Chat(voker_client, base_chat)

        if (
            base_responses := getattr(base_openai_client, "responses", None)
        ) is not None:
            self.responses = Responses(voker_client, base_responses)


class Chat:
    """Send calls to the chat endpoints to Voker."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_chat: openai_resources.chat.chat.Chat,
    ) -> None:
        self.__base_chat = base_chat

        if (base_completions := getattr(base_chat, "completions", None)) is not None:
            self.completions = ChatCompletions(voker_client, base_completions)

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying Chat object."""
        return getattr(self.__base_chat, name)


class ChatCompletions:
    """Send calls to the chat completion endpoints to Voker as llm events."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_chat_completions: openai_resources.chat.completions.completions.Completions,
    ) -> None:
        self._voker_client = voker_client
        self.__base_chat_completions = base_chat_completions

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying ChatCompletions object."""
        return getattr(self.__base_chat_completions, name)

    # TODO: how do i make this show up in intellisense properly?
    def create(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> (
        openai_types_chat.chat_completion.ChatCompletion
        | openai_streaming.Stream[
            openai_types_chat.chat_completion_chunk.ChatCompletionChunk
        ]
    ):
        """Send a chat completion request to Voker as an llm event.

        See `openai.OpenAI.chat.completions.create`
        """
        # TODO: add our other inputs in like person_id, ....
        # TODO: start time, end time, etc... # retries...
        to_stream = kwargs.get("stream", False)

        if not to_stream:
            response = self.__base_chat_completions.create(**kwargs)

            self._voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "openai-chat-completions",
                    "inputs": _rm_not_given(kwargs),
                    "output": response.model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

            return response

        def gen() -> (
            Generator[
                openai_types_chat.chat_completion_chunk.ChatCompletionChunk, None, None
            ]
        ):
            api_request: partial[
                openai_streaming.Stream[
                    openai_types_chat.chat_completion_chunk.ChatCompletionChunk
                ]
            ] = partial(self.__base_chat_completions.create, **kwargs)

            manager = openai.lib.streaming.chat.ChatCompletionStreamManager(
                api_request,
                response_format=kwargs.get("response_format", openai._types.omit),
                input_tools=kwargs.get("tools", openai._types.omit),
            )

            with manager as stream:
                for chunk in stream:
                    c = getattr(chunk, "chunk", None)

                    if c is None or len(c.choices) == 0:
                        continue

                    yield c

                response = stream.get_final_completion()

            self._voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "openai-chat-completions",
                    "inputs": _rm_not_given(kwargs),
                    "output": response.model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

        return gen()  # type: ignore


class Responses:
    """Send calls to the responses API to Voker."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_responses: openai_responses.responses.Responses,
    ) -> None:
        self._voker_client = voker_client
        self.__base_responses = base_responses

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying Responses object."""
        return getattr(self.__base_responses, name)

    def create(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> (
        openai_types_responses.response.Response
        | openai_streaming.Stream[
            openai_types_responses.response_stream_event.ResponseStreamEvent
        ]
    ):
        """Send a response request to Voker as an llm event.

        See `openai.OpenAI.responses.create`
        """
        # TODO: add our other inputs in like person_id, ....
        # TODO: start time, end time, etc... # retries...
        # TODO: instructions, conversation, previous_response_id, prompt, etc...
        to_stream = kwargs.get("stream", False)

        if not to_stream:
            response = self.__base_responses.create(**kwargs)

            self._voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "openai-responses",
                    "inputs": _rm_not_given(kwargs),
                    "output": response.model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

            return response

        from openai.lib.streaming.responses._responses import ResponseStreamManager

        def gen() -> Generator[
            openai_types_responses.response_stream_event.ResponseStreamEvent,
            None,
            None,
        ]:
            api_request: partial[
                openai_streaming.Stream[
                    openai_types_responses.response_stream_event.ResponseStreamEvent
                ]
            ] = partial(self.__base_responses.create, **kwargs)

            manager = ResponseStreamManager(
                api_request,
                text_format=kwargs.get("text_format", openai._types.omit),
                input_tools=kwargs.get("tools", openai._types.omit),
                starting_after=kwargs.get("starting_after", None),
            )

            with manager as stream:
                for chunk in stream:
                    yield chunk

                response = stream.get_final_response()

            self._voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "openai-responses",
                    "inputs": _rm_not_given(kwargs),
                    "output": response.model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

        return gen()  # type: ignore


class AsyncOpenAI(openai.AsyncOpenAI):
    """OpenAI wrapper that automatically creates llm events in Voker."""

    @overload
    def __init__(
        self,
        voker_client: Optional[VokerClient] = None,
        base_openai_client: Optional[openai.AsyncOpenAI] = None,
        *,
        api_key: Optional[str | Callable[[], Awaitable[str]]] = None,
        organization: Optional[str] = None,
        project: Optional[str] = None,
        webhook_secret: Optional[str] = None,
        base_url: Optional[str | httpx.URL] = None,
        websocket_base_url: Optional[str | httpx.URL] = None,
        timeout: Optional[float | openai.Timeout | openai.NotGiven] = openai.not_given,
        max_retries: int = openai.DEFAULT_MAX_RETRIES,
        default_headers: Optional[Mapping[str, str]] = None,
        default_query: Optional[Mapping[str, object]] = None,
        # Configure a custom httpx client.
        # We provide a `DefaultHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        http_client: Optional[httpx.AsyncClient] = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
        **kwargs,
    ) -> None: ...

    def __init__(
        self,
        voker_client: Optional[VokerClient] = None,
        base_openai_client: Optional[openai.AsyncOpenAI] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if voker_client is None:
            voker_client = VokerClient()

        if base_openai_client is None:
            base_openai_client = self

        # keep for testing access:
        self.__base_openai_client = base_openai_client
        self.__voker_client = voker_client

        if (base_chat := getattr(base_openai_client, "chat", None)) is not None:
            self.chat = AsyncChat(voker_client, base_chat)

        if (
            base_responses := getattr(base_openai_client, "responses", None)
        ) is not None:
            self.responses = AsyncResponses(voker_client, base_responses)


class AsyncChat:
    """Send calls to the chat endpoints to Voker as llm events."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_chat: openai_resources.chat.chat.AsyncChat,
    ) -> None:
        self.__base_chat = base_chat

        if (base_completions := getattr(base_chat, "completions", None)) is not None:
            self.completions = AsyncChatCompletions(voker_client, base_completions)

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying Chat object."""
        return getattr(self.__base_chat, name)


class AsyncChatCompletions:
    """Send calls to the chat completion endpoints to Voker as llm events."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_chat_completions: openai_resources.chat.completions.completions.AsyncCompletions,
    ) -> None:
        self._voker_client = voker_client
        self.__base_chat_completions = base_chat_completions

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying ChatCompletions object."""
        return getattr(self.__base_chat_completions, name)

    async def create(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> (
        openai_types_chat.chat_completion.ChatCompletion
        | openai_streaming.AsyncStream[
            openai_types_chat.chat_completion_chunk.ChatCompletionChunk
        ]
    ):
        """Send a chat completion request to Voker as an llm event.

        See `openai.OpenAI.chat.completions.create`
        """
        # TODO: add our other inputs in like person_id, ....
        # TODO: start time, end time, etc... # retries...

        to_stream = kwargs.get("stream", False)

        if not to_stream:
            response = await self.__base_chat_completions.create(**kwargs)

            self._voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "openai-chat-completions",
                    "inputs": _rm_not_given(kwargs),
                    "output": response.model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

            return response

        api_request = self.__base_chat_completions.create(**kwargs)

        manager = openai.lib.streaming.chat.AsyncChatCompletionStreamManager(
            api_request,
            response_format=kwargs.get("response_format", openai._types.omit),
            input_tools=kwargs.get("tools", openai._types.omit),
        )

        async def gen():
            async with manager as stream:
                async for chunk in stream:
                    c = getattr(chunk, "chunk", None)

                    if c is None or len(c.choices) == 0:
                        continue

                    yield c

                response = await stream.get_final_completion()

                self._voker_client.events.create(
                    event_name="llm",
                    properties={
                        "api": "openai-chat-completions",
                        "inputs": _rm_not_given(kwargs),
                        "output": response.model_dump(mode="json"),
                    },
                    agent=agent,
                    agent_version=agent_version,
                    person=person,
                    session=session,
                )

        return gen()


class AsyncResponses:
    """Send calls to the responses API to Voker as llm events."""

    def __init__(
        self,
        voker_client: VokerClient,
        base_responses: openai_responses.responses.AsyncResponses,
    ) -> None:
        self._voker_client = voker_client
        self.__base_responses = base_responses

    def __getattr__(self, name: str) -> Any:
        """Pass through to the underlying Responses object."""
        return getattr(self.__base_responses, name)

    async def create(
        self,
        agent: Optional[str] = None,
        agent_version: Optional[str] = None,
        person: Optional[str] = None,
        session: Optional[str] = None,
        **kwargs,
    ) -> (
        openai_types_responses.response.Response
        | openai_streaming.AsyncStream[
            openai_types_responses.response_stream_event.ResponseStreamEvent
        ]
    ):
        """Send a response request to Voker as an llm event.

        See `openai.OpenAI.responses.create`
        """
        # TODO: add our other inputs in like person_id, ....
        # TODO: start time, end time, etc... # retries...
        # TODO: instructions, conversation, previous_response_id, prompt, etc...
        to_stream = kwargs.get("stream", False)

        if not to_stream:
            response = await self.__base_responses.create(**kwargs)

            self._voker_client.events.create(
                event_name="llm",
                properties={
                    "api": "openai-responses",
                    "inputs": _rm_not_given(kwargs),
                    "output": response.model_dump(mode="json"),
                },
                agent=agent,
                agent_version=agent_version,
                person=person,
                session=session,
            )

            return response

        api_request = self.__base_responses.create(**kwargs)

        from openai.lib.streaming.responses._responses import AsyncResponseStreamManager

        manager = AsyncResponseStreamManager(
            api_request,
            text_format=kwargs.get("text_format", openai._types.omit),
            input_tools=kwargs.get("tools", openai._types.omit),
            starting_after=kwargs.get("starting_after", None),
        )

        async def gen():
            async with manager as stream:
                async for chunk in stream:
                    yield chunk

                response = await stream.get_final_response()

                self._voker_client.events.create(
                    event_name="llm",
                    properties={
                        "api": "openai-responses",
                        "inputs": _rm_not_given(kwargs),
                        "output": response.model_dump(mode="json"),
                    },
                    agent=agent,
                    agent_version=agent_version,
                    person=person,
                    session=session,
                )

        return gen()



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Event related payloads and responses."""

from typing import NotRequired, Required, TypedDict


class ClientCreateEventPayload(TypedDict):
    """Create an event.

    See [Creating an Event](https://app.voker.ai/docs/concepts/events#creating-an-event-via-sdk) for more information.
    """

    event_name: Required[str]
    properties: Required[EventProperties]

    agent: NotRequired[str]
    """The name of the [agent](https://app.voker.ai/docs/concepts/agents#agents) associated with the event.

    If this agent does not exist, it will be created. If an agent is not supplied, a default agent will be used.
    """

    agent_version: NotRequired[str]
    """The name of the [agent version](https://app.voker.ai/docs/concepts/agents#agent-versions) associated with the event.

    If this agent version does not exist, it will be created. If an agent version is not supplied, a default version will be used.
    """

    person: NotRequired[str]
    """The unique identifier of the [person](https://app.voker.ai/docs/concepts/people) associated with the event.

    If the person does not exist, it will be created.
    """

    session: NotRequired[str]
    """The unique identifier of the [session](https://app.voker.ai/docs/concepts/events#event-sessions) associated with the event.

    If the session does not exist, it will be created.
    """


class CreateEventPayload(ClientCreateEventPayload):
    """Create an event.

    See [Creating an Event](https://app.voker.ai/docs/concepts/events#creating-an-event-via-sdk) for more information.
    """

    fingerprint_id: Required[str]
    """The [fingerprint](https://app.voker.ai/docs/concepts/events#fingerprints) of the client that generated the event.
"""


EventProperties = dict



# Copyright 2026 Invoke Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Agent related payloads and responses."""

from typing import NotRequired, Required, TypedDict


class CreateAgentPayload(TypedDict):
    """Create an agent.

    See [Creating an Agent](https://app.voker.ai/docs/concepts/agents#creating-agent) for more information.
    """

    agent_name: Required[str]
    """Unique identifier for the agent."""

    description: NotRequired[str]
